
#include <stdio.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/keysym.h>
#include <GL/gl.h>
#include <GL/glx.h>
#include <stdlib.h>
#include <ctype.h>
#include <assert.h>
#include "glxoutput.h"
#include "vsync.h"

extern int glx_use_texture_mapping;
extern int pixel_format_bgra;

static Atom wmProtocolsAtom;
static Atom wmDeleteAtom;
static char *atomNames[] = { "WM_PROTOCOLS", "WM_DELETE_WINDOW" };

static Display *display;
static int screen;
static Window wnd;

static GLXContext ctx;

static int output_width, output_height;

static GLuint texture;
static GLubyte *pixels;
static int tex_width, tex_height;

static void gl_enter_2d_mode( int width, int height )
{
    glDisable( GL_DEPTH_TEST );
    glDepthMask( GL_FALSE );
    glDisable( GL_CULL_FACE );
    glDisable( GL_LIGHTING );
    glHint( GL_PERSPECTIVE_CORRECTION_HINT, GL_FASTEST );

    /* Set the view port to the size of our output. */
    glViewport( 0, 0, width, height );

    /* Setup 2D orthographic projection matrix. */
    glMatrixMode( GL_PROJECTION );
    glLoadIdentity();
    glOrtho( 0.0, width, height, 0.0, -1.0, 1.0 );

    glMatrixMode( GL_MODELVIEW );
    glLoadIdentity();
}


/* Stupid function for next highest power of two. */
static int power_of_two( int input )
{
    int value = 1;
    while( value < input ) value <<= 1;
    return value;
}

int glxoutput_init( int width, int height )
{
    char *window_name = "Text crawl viewer";
    char *icon_name = "Text crawl viewer";
    XSizeHints size_hints;
    XWindowAttributes attribs;
    XVisualInfo *vinfo;
    int glxAttrib[] = { GLX_RGBA, GLX_DOUBLEBUFFER, GLX_RED_SIZE, 8, GLX_GREEN_SIZE, 8, GLX_BLUE_SIZE, 8, None };

    /* Our intended output size. */
    output_width = width;
    output_height = height;

    /* Try texture mapping. */

    display = XOpenDisplay( 0 );
    if( !display ) {
        fprintf( stderr, "glxoutput: Can't open display '%s'.\n",
                 getenv( "DISPLAY" ) ? getenv( "DISPLAY" ) : "" );
        return 0;
    }
    screen = DefaultScreen( display );

    if( !vsync_init() ) {
        fprintf( stderr, "glxoutput: No retrace sync available.\n" );
        return 0;
    }

    size_hints.flags = PPosition | PSize | PMinSize | PMaxSize;
    size_hints.max_width = output_width;
    size_hints.min_width = output_width;
    size_hints.max_height = output_height;
    size_hints.min_height = output_height;

    /* Create the window. */
    wnd = XCreateSimpleWindow( display, XRootWindow( display, screen ), 0, 0,
                               output_width, output_height, 0,
                               BlackPixel( display, screen ), BlackPixel( display, screen ) );

    /* Set input stuff. */
    XSelectInput( display, wnd, ButtonPressMask | StructureNotifyMask |
                  KeyPressMask | ExposureMask | PropertyChangeMask );

    /* Set the properties. */
    XSetStandardProperties( display, wnd, window_name, icon_name, 0, 0, 0, &size_hints );
    XGetWindowAttributes( display, wnd, &attribs );

    /* Map it on the screen. */
    XMapWindow( display, wnd );

    /* Collaborate with the window manager for close requests. */
    XInternAtoms( display, atomNames, 2, False, &wmProtocolsAtom );
    XSetWMProtocols( display, wnd, &wmDeleteAtom, 1 );

    /* Setup the desired visual for OpenGL. */
    vinfo = glXChooseVisual( display, screen, glxAttrib );
    if( !vinfo ) {
        fprintf( stderr, "glxoutput: No acceptable GL visual available.\n" );
        return 0;
    }
    fprintf( stderr, "glxoutput: GL visual has red mask 0x%lx, green mask 0x%lx, blue mask 0x%lx\n",
             vinfo->red_mask, vinfo->green_mask, vinfo->blue_mask );

    ctx = glXCreateContext( display, vinfo, 0, True );
    if( !ctx ) {
        fprintf( stderr, "glxoutput: Can't create OpenGL context.\n" );
        return 0;
    }

    if( !glXMakeCurrent( display, wnd, ctx ) ) {
        fprintf( stderr, "glxoutput: Unable to attach OpenGL to our window!\n");
        return 0;
    }

    /* OpenGL is now active.  Enter 2D mode. */
    gl_enter_2d_mode( output_width, output_height );

    /* For now, we don't need alpha blending. */
    glDisable( GL_BLEND );
    glShadeModel( GL_FLAT );
    glBlendFunc( GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA );

    /* Try and get 8-byte alignment. */
    glPixelStorei( GL_UNPACK_ALIGNMENT, 8 );

    /* Stupid power-of-two sized textures. */
    tex_width = power_of_two( output_width );
    tex_height = power_of_two( output_height );
    fprintf( stderr, "glxoutput: Texture width: %d, texture height %d\n", tex_width, tex_height );

    /* Create the texture and the data. */
    pixels = malloc( tex_width * tex_height * 4 );

    /* Clear the framebuffer. */
    glClearColor( 0.0f, 0.0f, 0.0f, 0.0f );
    glClear( GL_COLOR_BUFFER_BIT );

     // memset( pixels, 0xff, tex_width * tex_height * 4 );
    if( glx_use_texture_mapping ) {
        glEnable( GL_TEXTURE_2D );
        fprintf( stderr, "glxoutput: Generating texture.\n" );

        glGenTextures( 1, &texture );    
        assert( glGetError() == GL_NO_ERROR );

        glBindTexture( GL_TEXTURE_2D, texture );
        assert( glGetError() == GL_NO_ERROR );

        glTexImage2D( GL_TEXTURE_2D, 0, GL_RGBA, tex_width, tex_height, 0,
                      GL_RGBA, GL_UNSIGNED_BYTE, 0 );
        assert( glGetError() == GL_NO_ERROR );

        glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT );
        glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT );
        glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR );
        glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR );
        glDisable( GL_TEXTURE_2D );
    }

    /* Display. */
    glXSwapBuffers( display, wnd );

    return 1;
}

void glxoutput_shutdown( void )
{
    vsync_shutdown();
}

void glxoutput_lock_output_buffer( void )
{
}

unsigned char *glxoutput_get_output_buffer( void )
{
    return pixels;
}

int glxoutput_get_output_stride( void )
{
    return ( output_width * 4 );
}

void glxoutput_unlock_output_buffer( void )
{
}

void glxoutput_show_frame( void )
{
    if( glx_use_texture_mapping ) {
        /**
         *  This code works on everything but the X4.3 mga driver?
         */
        double wtex = (double) output_width / (double) tex_width;
        double htex = (double) output_height / (double) tex_height;

        glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );
        glEnable( GL_TEXTURE_2D );
        glBindTexture( GL_TEXTURE_2D, texture );
        assert( glGetError() == GL_NO_ERROR );

        glColor3f( 1.0, 1.0, 1.0 );

        glTexSubImage2D( GL_TEXTURE_2D, 0, 0, 0, output_width, output_height,
                         pixel_format_bgra ? GL_BGRA : GL_RGBA, GL_UNSIGNED_BYTE, pixels );
        assert( glGetError() == GL_NO_ERROR );

        glMatrixMode( GL_MODELVIEW );
        glLoadIdentity();

        glBegin( GL_QUADS );
        glTexCoord2f( 0.0f, 0.0f ); glVertex2i( 0, 0 );
        glTexCoord2f( 0.0f, htex ); glVertex2i( 0, output_height );
        glTexCoord2f( wtex, htex ); glVertex2i( output_width, output_height );
        glTexCoord2f( wtex, 0.0f ); glVertex2i( output_width, 0 );
        glEnd();
        glDisable( GL_TEXTURE_2D );
    } else {
        glPixelZoom( 1.0, -1.0 );
        glRasterPos2i( 0, 0 );
        glDrawPixels( output_width, output_height, pixel_format_bgra ? GL_BGRA : GL_RGBA,
                      GL_UNSIGNED_BYTE, pixels );
    }

    /* Display the result. */
    glXSwapBuffers( display, wnd );
}

void glxoutput_wait_for_retrace( void )
{
    vsync_wait_for_retrace();
}

void glxoutput_set_window_caption( const char *cpation )
{
}

void glxoutput_poll_events( input_t *in )
{
    XEvent event;
    KeySym mykey;
    int arg = 0;
    char mykey_string;
    int reconfwidth = 0;
    int reconfheight = 0;

    while( XPending( display ) ) {
        XNextEvent( display, &event );

        switch( event.type ) {
        case ClientMessage:
            /* X sucks this way.  Thanks to walken for help on this one. */
            if( ( event.xclient.message_type == wmProtocolsAtom ) && ( event.xclient.format == 32 ) &&
                ( event.xclient.data.l[ 0 ] == wmDeleteAtom ) ) {
                input_callback( in, I_QUIT, 0 );
            }
            break;
        case DestroyNotify:
            if( event.xdestroywindow.window != wnd ) {
                break;
            }
            input_callback( in, I_QUIT, 0 );
            break;
        case Expose:
            break;
        case ConfigureNotify:
            reconfwidth = event.xconfigure.width;
            reconfheight = event.xconfigure.height;
            break;
        case KeyPress:
            mykey = XKeycodeToKeysym( display, event.xkey.keycode, 0 );
            if( event.xkey.state & ShiftMask ) arg |= I_SHIFT;
            if( event.xkey.state & ControlMask ) arg |= I_CTRL;
            if( event.xkey.state & Mod1Mask ) arg |= I_META;

            if( mykey >= XK_a && mykey <= XK_z ) {
                arg |= mykey - XK_a + 'a';
            } else if( mykey >= XK_A && mykey <= XK_Z ) {
                arg |= mykey - XK_A + 'a';
            } else if( mykey >= XK_0 && mykey <= XK_9 ) {
                arg |= mykey - XK_0 + '0';
            } else {
                XLookupString( &event.xkey, &mykey_string, 1, 0, 0 );
                arg |= mykey_string;
            }

            switch( mykey ) {
            case XK_Escape: arg |= I_ESCAPE; break;

            case XK_Home:      arg |= I_HOME; break;
            case XK_End:       arg |= I_END; break;
            case XK_Left:      arg |= I_LEFT; break;
            case XK_Up:        arg |= I_UP; break;
            case XK_Right:     arg |= I_RIGHT; break;
            case XK_Down:      arg |= I_DOWN; break;
            case XK_Page_Up:   arg |= I_PGUP; break;
            case XK_Page_Down: arg |= I_PGDN; break;

            case XK_F1: arg |= I_F1; break;
            case XK_F2: arg |= I_F2; break;
            case XK_F3: arg |= I_F3; break;
            case XK_F4: arg |= I_F4; break;
            case XK_F5: arg |= I_F5; break;
            case XK_F6: arg |= I_F6; break;
            case XK_F7: arg |= I_F7; break;
            case XK_F8: arg |= I_F8; break;
            case XK_F9: arg |= I_F9; break;
            case XK_F10: arg |= I_F10; break;
            case XK_F11: arg |= I_F11; break;
            case XK_F12: arg |= I_F12; break;

            case XK_KP_Enter: arg |= I_ENTER; break;
            case XK_KP_0: arg |= '0'; break;
            case XK_KP_1: arg |= '1'; break;
            case XK_KP_2: arg |= '2'; break;
            case XK_KP_3: arg |= '3'; break;
            case XK_KP_4: arg |= '4'; break;
            case XK_KP_5: arg |= '5'; break;
            case XK_KP_6: arg |= '6'; break;
            case XK_KP_7: arg |= '7'; break;
            case XK_KP_8: arg |= '8'; break;
            case XK_KP_9: arg |= '9'; break;
            case XK_KP_Multiply: arg |= '*'; break;
            case XK_KP_Subtract: arg |= '-'; break;
            case XK_KP_Add: arg |= '+'; break;
            case XK_KP_Divide: arg |= '/'; break;

            case XK_KP_End: arg |= '1'; break;
            case XK_KP_Down: arg |= '2'; break;
            case XK_KP_Next: arg |= '3'; break;
            case XK_KP_Left: arg |= '4'; break;
            case XK_KP_Begin: arg |= '5'; break;
            case XK_KP_Right: arg |= '6'; break;
            case XK_KP_Home: arg |= '7'; break;
            case XK_KP_Up: arg |= '8'; break;
            case XK_KP_Prior: arg |= '9'; break;

            default: break;
            }
            input_callback( in, I_KEYDOWN, arg );
            break;
        case ButtonPress:
            input_callback( in, I_BUTTONPRESS, event.xbutton.button );
            break;
        default: break;
        }
    }
}

output_api_t glxoutput_api = {
    glxoutput_init,
    glxoutput_shutdown,

    glxoutput_lock_output_buffer,
    glxoutput_get_output_buffer,
    glxoutput_get_output_stride,
    glxoutput_unlock_output_buffer,

    glxoutput_show_frame,
    glxoutput_wait_for_retrace,
    glxoutput_set_window_caption,
    glxoutput_poll_events
};

output_api_t *glxoutput_get_output( void )
{
    return &glxoutput_api;
}

